@@ -1,6 +1,6 @@
 package App::Greple::Grep;
 
-use v5.14;
+use v5.24;
 use warnings;
 
 use Exporter 'import';
@@ -12,6 +12,7 @@ our @ISA = qw(App::Greple::Text);
 
 use Data::Dumper;
 use List::Util qw(min max reduce sum);
+use Clone qw(clone);
 
 use Getopt::EX::Func qw(callable);
 
@@ -62,6 +63,30 @@ sub run {
     $self->prepare->compose;
 }
 
+package App::Greple::Grep::Block {
+    use strict;
+    sub min    :lvalue { $_[0]->[0] }
+    sub max    :lvalue { $_[0]->[1] }
+    sub number :lvalue { $_[0]->[2] }
+}
+
+package App::Greple::Grep::Match {
+    use strict;
+    sub min      :lvalue { $_[0]->[0] }
+    sub max      :lvalue { $_[0]->[1] }
+    sub index    :lvalue { $_[0]->[2] }
+    sub callback :lvalue { $_[0]->[3] }
+}
+
+package App::Greple::Grep::Result {
+    use strict;
+    sub block   { $_[0]->[0] }
+    sub matched { $_[0]->@[ 1 .. $#{$_[0]} ] }
+    sub min     { $_[0]->block->min }
+    sub max     { $_[0]->block->max }
+    sub number  { $_[0]->block->number }
+}
+
 sub prepare {
     my $self = shift;
 
@@ -77,6 +102,7 @@ sub prepare {
     ##
     my @result;
     my $positive_count = 0;
+    my $group_index_offset = 0;
     my @patlist = $pat_holder->patterns;
     while (my($i, $pat) = each @patlist) {
 	my($func, @args) = do {
@@ -85,8 +111,8 @@ sub prepare {
 	    } else {
 		Getopt::EX::Func->new(\&match_regions,
 				      pattern => $pat->regex,
-				      group => $self->{group_match},
-				      index => $self->{group_index},
+				      group => $self->{group_index},
+				      index => $self->{group_index} >= 2,
 		    );
 	    }
 	};
@@ -97,19 +123,43 @@ sub prepare {
 	    ## required pattern can be compromised upto that number.
 	    ##
 	    return $self if $pat->is_required and $self->{need} >= 0;
+	    ##
+	    ## Update offset even when no match for --ci=G
+	    ##
+	    if ($self->{group_index} == 2 and not $pat->is_function) {
+		$group_index_offset += $pat->group_count //= do {
+		    "" =~ /@{[$pat->regex]}?/;
+		    $#+ || 1;
+		};
+	    }
 	} else {
+	    bless $_, 'App::Greple::Grep::Match' for @p;
 	    if ($pat->is_positive) {
-		push @blocks, map { [ @$_ ] } @p;
+		push @blocks, @{clone(\@p)};
 		$self->{stat}->{match_positive} += @p;
 		$positive_count++;
 	    }
 	    else {
 		$self->{stat}->{match_negative} += @p;
 	    }
-	    map { $_->[2] //= $i } @p;
-	    if (my $n = @{$self->{callback}}) {
+	    ##
+	    ## Adjust group index for --ci=G option
+	    ## group_index: 0=off, 1=group, 2=sequential, 3=per-pattern
+	    ##
+	    if ($self->{group_index} == 2) {
+		my $max_index = 0;
+		for (@p) {
+		    if (defined $_->index) {
+			$_->index += $group_index_offset;
+			$max_index = $_->index + 1 if $_->index >= $max_index;
+		    }
+		}
+		$group_index_offset = $max_index;
+	    }
+	    $_->index //= $i for @p;
+	    if (my $n = $self->{callback}->@*) {
 		if (my $callback = $self->{callback}->[ $i % $n ]) {
-		    map { $_->[3] //= $callback } @p;
+		    $_->callback //= $callback for @p;
 		}
 	    }
 	}
@@ -158,12 +208,11 @@ sub prepare {
     ## Setup BLOCKS
     ##
     my $bp = $self->{BLOCKS} = [ do {
-	if (@{$self->{block}}) {		# --block
+	if ($self->{block}->@*) {		# --block
 	    my $text = \$_;
 	    merge_regions { nojoin => 1, destructive => 1 }, map {
-		grep { $_->[0] != $_->[1] }
 		get_regions($self->{filename}, $text, $_);
-	    } @{$self->{block}};
+	    } $self->{block}->@*;
 	}
 	elsif (@blocks) {			# from matched range
 	    my %opt = ( A => $self->{after},
@@ -171,13 +220,18 @@ sub prepare {
 		        border => [ $self->borders ] );
 	    my $blocker = smart_blocker(\%opt);
 	    merge_regions { nojoin => 1, destructive => 1 }, map {
-		[ $blocker->(\%opt, $_->[0], $_->[1]) ]
+		[ $blocker->(\%opt, $_->min, $_->max) ]
 	    } @blocks;
 	}
 	else {
 	    ( [ 0, length ] );			# nothing matched
 	}
     } ];
+    while (my($i, $blk) = each @$bp) {
+	bless $blk, 'App::Greple::Grep::Block';
+	# set 1-origined block number in the 3rd entry
+	$blk->number = $i + 1;
+    }
 
     ##
     ## build match table
@@ -225,7 +279,7 @@ sub compose {
     ##
     if (my $countcheck = $self->{countcheck}) {
 	@effective_index = do {
-	    grep { $countcheck->(int(@{$mp->[$_][POSI_LIST]})) }
+	    grep { $countcheck->(int($mp->[$_][POSI_LIST]->@*)) }
 	    @effective_index;
 	}
 	or return $self;
@@ -234,7 +288,7 @@ sub compose {
     ##
     ## --block with -ABC option
     ##
-    if (@{$self->{block}} and ($self->{after} or $self->{before})) {
+    if ($self->{block}->@* and ($self->{after} or $self->{before})) {
 	my @mark;
 	for my $i (@effective_index) {
 	    map { $mark[$_] = 1 if $_ >= 0 }
@@ -249,14 +303,14 @@ sub compose {
     my @list = ();
     for my $bi (@effective_index) {
 	my @matched = merge_regions({ nojoin => 1, destructive => 1 },
-				    @{$mp->[$bi][MUST_LIST]},
-				    @{$mp->[$bi][POSI_LIST]},
-				    @{$mp->[$bi][NEGA_LIST]});
+				    $mp->[$bi][MUST_LIST]->@*,
+				    $mp->[$bi][POSI_LIST]->@*,
+				    $mp->[$bi][NEGA_LIST]->@*);
 	if ($self->{stretch}) {
 	    my $b = $bp->[$bi];
 	    my $m = $matched[0];
 	    my $i = min map { $_->[2] // 0 } @matched;
-	    @matched = [ $b->[0], $b->[1], $i, $m->[3] ];
+	    @matched = [ $b->min, $b->max, $i, $m->[3] ];
 	}
 	if ($self->{only}) {
 	    push @list, map({ [ $_, $_ ] } @matched);
@@ -267,14 +321,21 @@ sub compose {
 	    push @list, [ $bp->[$bi], @matched ];
 	}
     }
+    for my $r (@list) {
+	bless $r, 'App::Greple::Grep::Result';
+	bless $r->block, 'App::Greple::Grep::Block';
+	for my $m ($r->matched) {
+	    bless $m, 'App::Greple::Grep::Match';
+	}
+    }
 
     ##
     ## --join-blocks
     ##
     if ($self->{join_blocks} and @list > 1) {
 	reduce {
-	    if ($a->[-1][0][-1] == $b->[0][0]) {
-		$a->[-1][0][-1]  = $b->[0][1];
+	    if ($a->[-1][0]->max == $b->[0]->min) {
+		$a->[-1][0]->max  = $b->[0]->max;
 		push @{$a->[-1]}, splice @$b, 1;
 	    } else {
 		push @$a, $b;
@@ -284,8 +345,8 @@ sub compose {
     }
 
     ##
-    ## ( [ [blockstart, blockend], [start, end], [start, end], ... ],
-    ##   [ [blockstart, blockend], [start, end], [start, end], ... ], ... )
+    ## ( [ [blockstart, blockend, number ], [start, end], [start, end], ... ],
+    ##   [ [blockstart, blockend, number ], [start, end], [start, end], ... ], ... )
     ##
     $self->{RESULT} = \@list;
 
@@ -333,19 +394,19 @@ sub result {
 
 sub matched {
     my $obj = shift;
-    sum(map { @{$_} - 1 } $obj->result) // 0;
+    sum(map { $_->@* - 1 } $obj->result) // 0;
 }
 
 sub blocks {
     my $obj = shift;
-    @{ $obj->{BLOCKS} };
+    $obj->{BLOCKS}->@*;
 }
 
 sub slice_result {
     my $obj = shift;
     my $result = shift;
     my($block, @list) = @$result;
-    my $template = unpack_template(\@list, $block->[0]);
+    my $template = unpack_template(\@list, $block->min);
     unpack($template, $obj->cut(@$block));
 }
 
@@ -424,39 +485,205 @@ sub blocker {
     @$border[ $bi, $ei ];
 }
 
-1;
-
-
-package App::Greple::Text;
-
-use strict;
-use warnings;
-
-use overload 
-    '""' => \&stringify;
-
-sub stringify {
-    my $obj = shift;
-    ${ $obj->{text} };
-}
-
-sub new {
-    my $class = shift;
-    my $obj = bless {
-	text => \$_[0],
-    }, $class;
-    $obj;
-}
-
-sub text {
-    my $obj = shift;
-    ${ $obj->{text} };
-}
-
-sub cut {
-    my $obj = shift;
-    my($from, $to) = @_;
-    substr ${ $obj->{text} }, $from, $to - $from;
+package App::Greple::Text {
+    use strict;
+    use warnings;
+    use overload '""' => sub { ${ $_[0]->{text} } };
+    sub new {
+	my $class = shift;
+	bless { text => \$_[0] }, $class;
+    }
+    sub text { ${ $_[0]->{text} } }
+    sub cut {
+	my($obj, $from, $to) = @_;
+	substr ${ $obj->{text} }, $from, $to - $from;
+    }
 }
 
 1;
+
+__END__
+
+=encoding utf-8
+
+=head1 NAME
+
+App::Greple::Grep - Greple grep engine module
+
+=head1 SYNOPSIS
+
+    use App::Greple::Grep;
+
+    my $grep = App::Greple::Grep->new(
+        text     => \$text,
+        filename => $filename,
+        pattern  => $pattern_holder,
+        ...
+    )->run;
+
+    for my $result ($grep->result) {
+        my $block = $result->block;
+        for my $match ($result->matched) {
+            # process each match
+        }
+    }
+
+=head1 DESCRIPTION
+
+This module provides the core grep engine for the B<greple> command.
+It is typically not used directly, but can be accessed through the
+C<--postgrep> option to manipulate search results.
+
+=head1 CLASSES
+
+=head2 App::Greple::Grep::Block
+
+Represents a text block containing matched patterns.
+
+=over 4
+
+=item B<min>
+
+Start position of the block (0-origin byte offset).
+
+=item B<max>
+
+End position of the block.
+
+=item B<number>
+
+Block number (1-origin).
+
+=back
+
+=head2 App::Greple::Grep::Match
+
+Represents an individual match within a block.
+
+=over 4
+
+=item B<min>
+
+Start position of the match.
+
+=item B<max>
+
+End position of the match.
+
+=item B<index>
+
+Pattern index (0-origin) indicating which pattern matched.
+Used for colorization with C<--colorindex> option.
+
+=item B<callback>
+
+Callback function to be called when this match is output.
+The callback receives the following arguments:
+
+For CODE reference:
+
+    $callback->($start, $end, $index, $matched_string)
+
+For L<Getopt::EX::Func> object:
+
+    $callback->call(
+        &FILELABEL => $file,
+        start => $start,
+        end   => $end,
+        index => $index,
+        match => $matched_string,
+    )
+
+The callback should return the string to be output.  If it returns
+C<undef>, the original matched string is used.
+
+=back
+
+=head2 App::Greple::Grep::Result
+
+Represents the search result for a single block.
+
+=over 4
+
+=item B<block>
+
+Returns the C<App::Greple::Grep::Block> object.
+
+=item B<matched>
+
+Returns a list of C<App::Greple::Grep::Match> objects.
+
+=item B<min>, B<max>, B<number>
+
+Shortcuts for C<< $result->block->min >>, etc.
+
+=back
+
+=head1 METHODS
+
+=head2 App::Greple::Grep
+
+=over 4
+
+=item B<result>
+
+Returns a list of C<App::Greple::Grep::Result> objects.
+
+=item B<result_ref>
+
+Returns a reference to the result array.  This can be used to modify
+the results in a C<--postgrep> function.
+
+=item B<matched>
+
+Returns the total number of matches.
+
+=item B<blocks>
+
+Returns a list of all blocks.
+
+=item B<cut>(I<$from>, I<$to>)
+
+Returns the substring of the text from position I<$from> to I<$to>.
+
+=item B<slice_result>(I<$result>)
+
+Returns a list of strings alternating between unmatched and matched
+portions within the block.
+
+=back
+
+=head1 USING --postgrep OPTION
+
+The C<--postgrep> option allows you to process the Grep object after
+the search is complete.  The function receives the Grep object as its
+argument.
+
+    sub postgrep {
+        my $grep = shift;
+        for my $result ($grep->result) {
+            for my $match ($result->matched) {
+                # Modify match attributes
+                $match->callback = sub { ... };
+            }
+        }
+    }
+
+=head1 CALLBACK EXAMPLE
+
+The L<App::Greple::subst> module uses callbacks to replace matched
+text:
+
+    my $callback = sub {
+        my($start, $end, $index, $matched) = @_;
+        # Return replacement string
+        return $replacement;
+    };
+
+    for my $match ($result->matched) {
+        $match->callback = $callback;
+    }
+
+=head1 SEE ALSO
+
+L<greple>, L<App::Greple::subst>
